/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://212.187.12.197/RNG/terraform/
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// hacked by Adam Rotaru <adam_rotaru@altavista.net>, Mar 2000
// this file adapted from an -lgtkglmm demo by 
// Karl Nelson <kenelson@ece.ucdavis.edu>


#include <stdio.h>

#include "TFOpenGLArea.h" 
#include "GlobalSanityCheck.h"
#include "MathTrig.h"
#include "TFGui.h"
#include "TFOptions.h"
#include "TFWindowHandler.h"		// circular




/** 
 * TFOpenGLArea::TFOpenGLArea
 * 	This constructor makes the connections necessary to
 * 	handle exposures and to initialize the GL object.
 * 
 *  Note: We can NOT constuct or execute any GL commands
 *  at this point because we have not yet been realized.
 *  Therefore, initialization of the GL odjects must be
 *  done in a seperate realization function.  In this
 *  case we will connect it to initGL().
 *
 *  (the show signal happens before realization and should 
 *   not be used.)
 */ 
TFOpenGLArea::TFOpenGLArea(const HeightField *HF, const HeightFieldDraw *HFD) 
	     : Gtk::GLArea() 
{
	SanityCheck::bailout ((!HF), "HF==NULL", "TFOpenGLArea::TFOpenGLArea");
	SanityCheck::bailout ((!HFD), "HFD==NULL", "TFOpenGLArea::TFOpenGLArea");

	// assign our real datapointers 
	p_HF = const_cast<HeightField*> (HF);
	p_HFD = const_cast<HeightFieldDraw*> (HFD);

	d_terrainList  = d_normalsList = 0;

	d_meshColumnStep = 4;
	d_meshRowStep = 4;
	d_doDrawNormals = GL_FALSE;

	p_meshVertices = NULL;
	p_meshNormals = NULL;

	//select events so that we can properly connect to them
	//GDK_POINTER_MOTION_HINT_MASK is used so that we 
	//don't get every point of motion
	this->set_events(GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
			 GDK_KEY_PRESS_MASK);
}


TFOpenGLArea::~TFOpenGLArea()
{
	for (int i = 0; i < p_HF->getWidth(); i++) 
		{
		for (int j = 0; j < p_HF->getHeight(); j++) 
			{
			delete [] p_meshVertices[i][j];
			delete [] p_meshNormals[i][j];
			}
		delete [] p_meshVertices[i];
		delete [] p_meshNormals[i];
		}

	delete [] p_meshVertices;
	delete [] p_meshNormals;

	if (glIsList(d_terrainList)==GL_TRUE)
		glDeleteLists(d_terrainList,1);
	if (glIsList(d_normalsList)==GL_TRUE)
		glDeleteLists(d_normalsList,1);
}



void TFOpenGLArea::realize_impl()
{
	// we must call the real realize_impl so the object is marked realized
	Gtk::GLArea::realize_impl();

	// then we can call our init
	initGL();
}


gint TFOpenGLArea::expose_event_impl (GdkEventExpose *event)
{
	// we are CPU intensive so we should wait for the last event
	if (event->count > 0) 
		return TRUE;

	drawGL();
	return TRUE;
}


gint TFOpenGLArea::configure_event_impl (GdkEventConfigure *event)
{
	if (make_current()) 
		{
      		resizeGL();
		return TRUE;
		}
	return FALSE;
}


/** 
 * TFOpenGLArea::initGL
 * 	This function creates the OpenGL list that will be used 
 * 	to draw the terrain. We cannot do any GL rendering until
 * 	the widget is realized including creating lists.
 *
 *	This should probably use a combination of RenderOptions, 
 * 	PlaceObjectOptions and maybe TFOptions (for the Y scale 
 * 	factor). This would allow us to have just one set of 
 * 	parameters.  
 */
void TFOpenGLArea::initGL()
{
	int 		i, j;
	int		limi, limj;
	PTYPE		elv1;

	// In order to allow the window to be shrunk we will 
	// reset the minumum size to something smaller now 
	// that the inital allocation has been completed.
	//  (this works, but is this what they intended??)
	set_usize(100,100);

	// make_current must be called before performing GL operations
	// and GL code should only be called if make_current returns true
	if (make_current())
		{
		setupGL();
        
                d_terrainList = glGenLists(1);
		if (!d_terrainList)
			return;

		buildMesh ();

		// store the drawing of the mesh in a display list
                glNewList(d_terrainList, GL_COMPILE);
		limi = p_HF->getWidth();
		limj = p_HF->getHeight()-d_meshRowStep;
                for(j = 0; j < limj; j += d_meshColumnStep) 
		    {
		    // draw a row as a triangle strip
		    // (a quad can have non-convex projection)
                    glBegin(GL_TRIANGLE_STRIP);
                    // glBegin(GL_QUADS); 
                    for(i = 0; i < limi; i += d_meshRowStep)
			{                
			const	GdkColor	*col;
			// for each vertex, set the color
			// (amb. and diffuse), the normal, and coordinates

 			elv1 = 0.1 + 0.8*p_HF->getEl(i,j+d_meshRowStep);
			glColor3f(elv1, 0.1, 0.1);
			// for some reason, using the colormap colors 
			// doesn't give a good result
			//col = p_HFD->getColor2d (i,j+d_meshRowStep);
			//glColor3f(col->red, col->green, col->blue);
			glNormal3fv(p_meshNormals[i][j+d_meshRowStep]); 
			glVertex3fv(p_meshVertices[i][j+d_meshRowStep]);

			elv1 = 0.1 + 0.8*p_HF->getEl(i,j); 
			glColor3f(elv1, 0.1, 0.1); 	
			// see above
			//col = p_HFD->getColor2d (i, j);
			//glColor3f(col->red, col->green, col->blue);
			glNormal3fv(p_meshNormals[i][j]); 	 
			glVertex3fv(p_meshVertices[i][j]);
			}   
                    glEnd();
		    }                
		glEndList();       

		// draw normals 
		d_normalsList = glGenLists(1);
		glNewList(d_normalsList, GL_COMPILE);
		glBegin(GL_LINES);
		for (i=0; i<limi; i+=d_meshColumnStep) 
		    {
		    for (j=0; j<limj; j+=d_meshRowStep) 
			{
			glColor3f(1.0, 0.0, 0.0);
			glVertex3fv(p_meshVertices[i][j]);
			glColor3f(1.0, 0.5, 0.5);
			glVertex3f(
			 p_meshVertices[i][j][0]+0.05*p_meshNormals[i][j][0],
			 p_meshVertices[i][j][1]+0.05*p_meshNormals[i][j][1],
			 p_meshVertices[i][j][2]+0.05*p_meshNormals[i][j][2]
			);
			}
		    }
		glEnd();
  		glEndList();
		//fflush (stdout);
		}
}


/**
 * precompute the 3D coordinates of the mesh points,
 * and normals
 */
void TFOpenGLArea::buildMesh ()
{
	int 		i, j;
	int		h,w;
	float		max;
	float		ys = TFOptions::s_yscale;
	GLfloat*** 	mesh_facetnormals;
	GLfloat 	u[3], v[3], n[3], l;

	w = p_HF->getWidth();
	h = p_HF->getHeight();
	max = p_HF->getMax ();

	// allocate space for the mesh 
	if (!p_meshVertices)
		{
		p_meshVertices	 	= new GLfloat**[w]; 
		p_meshNormals	 	= new GLfloat**[w]; 
		for (i = 0; i < w; i++) 
			{
			p_meshVertices[i]  	= new GLfloat*[h];
			p_meshNormals[i]    = new GLfloat*[h];
			for (j = 0; j < h; j++) 
				{
				p_meshVertices[i][j]	= new GLfloat[3];
				p_meshNormals[i][j]	= new GLfloat[3];
				}
			}
		}

	// allocate normals
	mesh_facetnormals = new GLfloat**[w]; 
	for (i = 0; i < w; i++) 
		{
		mesh_facetnormals[i] = new GLfloat*[h];
		for (j = 0; j < h; j++) 
			{
			mesh_facetnormals[i][j] = new GLfloat[3];
			}
		}

	// fill out the vertex array and calculate facet normals for the mesh
	for (i = 0; i < w - 1; i++) 
	    {
	    for (j = 0; j < h - 1; j++) 
		{
		// assign the data to vertices. Some of the vertices will be
		// overwritten in subsequent iterations of the loop, but this is
		// OK, since they will be identical. 
		p_meshVertices[i][j][0] = ((GLfloat)i / (GLfloat)w) - 0.5; 
		p_meshVertices[i][j][1] = (p_HF->getElSea(i,j)/max)*ys - (0.5*ys);
  		p_meshVertices[i][j][2] = ((GLfloat)j / (GLfloat)h) - 0.5;

		p_meshVertices[i][j+1][0] = p_meshVertices[i][j][0];
		p_meshVertices[i][j+1][1] = (p_HF->getElSea(i,j+1)/max)*ys - (0.5*ys);
		p_meshVertices[i][j+1][2] = ((GLfloat)(j+1) / (GLfloat)h) - 0.5;

		p_meshVertices[i+1][j][0] = ((GLfloat)(i+1) / (GLfloat)w)- 0.5;
		p_meshVertices[i+1][j][1] = (p_HF->getElSea(i+1,j)/max)*ys - (0.5*ys);
		p_meshVertices[i+1][j][2] = p_meshVertices[i][j][2];

		// get two vectors to cross 
		u[0] = p_meshVertices[i][j+1][0] - p_meshVertices[i][j][0];
		u[1] = p_meshVertices[i][j+1][1] - p_meshVertices[i][j][1];
		u[2] = p_meshVertices[i][j+1][2] - p_meshVertices[i][j][2];

		v[0] = p_meshVertices[i+1][j][0] - p_meshVertices[i][j][0];
		v[1] = p_meshVertices[i+1][j][1] - p_meshVertices[i][j][1];
		v[2] = p_meshVertices[i+1][j][2] - p_meshVertices[i][j][2];

		// get the normalized cross product 
		MathTrig::normalVectorNormalized (u, v, n);
      
		// put the facet normal in the i, j position for later 
		// averaging with other normals. 
		mesh_facetnormals[i][j][0] = n[0];
		mesh_facetnormals[i][j][1] = n[1];
		mesh_facetnormals[i][j][2] = n[2];
		}
	    }

	// fill in the last vertex & it's facet normal 
	p_meshVertices[i][j][0] = ((GLfloat)i / (GLfloat)w) - 0.5; 
	p_meshVertices[i][j][1] = (p_HF->getEl(i,j) / max) - 0.5;
	p_meshVertices[i][j][2] = ((GLfloat)j / (GLfloat)h) - 0.5;

	mesh_facetnormals[i][j][0] = n[0];
	mesh_facetnormals[i][j][1] = n[1];
	mesh_facetnormals[i][j][2] = n[2];

	// calculate normals for the mesh 
	for (i = 1; i < w - 1; i++) 
	    {
	    for (j = 1; j < h - 1; j++) 
		{
		// average all the neighboring normals. 
		n[0] = mesh_facetnormals[i-1][j-1][0];
		n[1] = mesh_facetnormals[i-1][j-1][1];
		n[2] = mesh_facetnormals[i-1][j-1][2];

		n[0] += mesh_facetnormals[i][j-1][0];
		n[1] += mesh_facetnormals[i][j-1][1];
		n[2] += mesh_facetnormals[i][j-1][2];

		n[0] += mesh_facetnormals[i-1][j][0];
		n[1] += mesh_facetnormals[i-1][j][1];
		n[2] += mesh_facetnormals[i-1][j][2];

		n[0] += mesh_facetnormals[i][j][0];
		n[1] += mesh_facetnormals[i][j][1];
		n[2] += mesh_facetnormals[i][j][2];
      
		l = (GLfloat)sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
		p_meshNormals[i][j][0] = n[0] /= l;
		p_meshNormals[i][j][1] = n[1] /= l;
		p_meshNormals[i][j][2] = n[2] /= l;
		}
	    }

	// fill in the normals on the top/bottom edge of the mesh (simply
	// copy the one below/above it). 
	for (i = 0; i < w; i++) 
		{
		p_meshNormals[i][0][0] = p_meshNormals[i][1][0];
		p_meshNormals[i][0][1] = p_meshNormals[i][1][1];
		p_meshNormals[i][0][2] = p_meshNormals[i][1][2];

		p_meshNormals[i][h-1][0] = p_meshNormals[i][h-2][0];
		p_meshNormals[i][h-1][1] = p_meshNormals[i][h-2][1];
		p_meshNormals[i][h-1][2] = p_meshNormals[i][h-2][2];
		}

	// fill in the normals on the left/right edge of the mesh (simply
	// copy the one right/left of it).
	for (j = 0; j < h; j++) 
		{
		p_meshNormals[0][j][0] = p_meshNormals[1][j][0];
		p_meshNormals[0][j][1] = p_meshNormals[1][j][1];
		p_meshNormals[0][j][2] = p_meshNormals[1][j][2];

		p_meshNormals[w-1][j][0] = p_meshNormals[w-2][j][0];
		p_meshNormals[w-1][j][1] = p_meshNormals[w-2][j][1];
		p_meshNormals[w-1][j][2] = p_meshNormals[w-2][j][2];
		}

	// printf("%d x %d x %.0f mesh\n", w, h, max);

	// free the mesh data & facet normals 
	for (i = 0; i < w; i++) 
		{
		for (j = 0; j < h; j++)
			delete mesh_facetnormals[i][j];
		delete [] mesh_facetnormals[i];
		}
	delete [] mesh_facetnormals;
}



/**
 * drawTerrain: draw the terrain object
 */
void TFOpenGLArea::drawTerrain () const
{
	glCallList(d_terrainList);

	if (d_doDrawNormals)
		glCallList(d_normalsList);
}


/**
 * resizeGL: Adjusts the projection matrix to the window size.
 * 	This is called once to set up the initial matrix and
 * 	then again if the window is resized.
 */
void TFOpenGLArea::resizeGL () const
{
	int		width, height;
	GLfloat		aspect;

	//fprintf(stderr, "resizeGL\n");
	width = this->width();
	height = this->height();
	glViewport(0, 0, width, height);
	// create the projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	aspect = ((float)width)/((float)height);
	if (width > height)
		glFrustum(-0.12*aspect, 0.12*aspect, -0.12, 0.12, 0.2, 20);
	else
		glFrustum(-0.12, 0.12, -0.12/aspect, 0.12/aspect, 0.2, 20);
	gluLookAt(0.25, 0.7, 0.8,		// eye pos
		0.0,	0.0,	0.0,		// looking at
		0.0,	1.0,	0.0);		// up vector
}


/**
 * setupGL: configures the OpenGL rendering engine
 *
 * This is called once.
 */
void TFOpenGLArea::setupGL() const
{
	static GLfloat	tmp[4];

	//fprintf(stderr, "setupGL\n");
	// Setup the rendering engine
	// frame buffer clears should be to black
	glClearColor (0.0, 0.0, 0.0, 1.0); 
  	glClearDepth(1.0);
	// Z-buffering enabled
	glEnable(GL_DEPTH_TEST);
	// draw only forward-face of polygons
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);
	// we supply normalized normals, no need to renormalize
	glDisable(GL_NORMALIZE);
	// if we draw lines (normals)
	glLineWidth(0.5f);
	// no line antialiasing
	glDisable(GL_LINE_SMOOTH);

	// materials
	// we want to be able to set the matierial ambient and diffuse
	// color by glColor().  (This is also the default value.)
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	// lights
	// amabient color (seen in shady areas - almost black)
	tmp[0] = 0.1; tmp[1] = 0.1; tmp[2] = 0.1; tmp[3] = 1.0;
	glLightfv (GL_LIGHT0, GL_AMBIENT, tmp);
	// diffuese color (*the* color of the light - white)
	tmp[0] = 0.9; tmp[1] = 1.0; tmp[2] = 0.9; tmp[3] = 1.0;
	glLightfv (GL_LIGHT0, GL_DIFFUSE, tmp);
	// it is directional light (4th value is 0)
	tmp[0] = 0.3; tmp[1] = 0.5; tmp[2] = -0.5; tmp[3] = 0.0;
	glLightfv (GL_LIGHT0, GL_POSITION, tmp);
	glEnable (GL_LIGHT0);
	// to be more precise, use local viewpoint
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glEnable (GL_LIGHTING);
	
	// fog
	glFogi(GL_FOG_MODE, GL_EXP);
	glFogf(GL_FOG_DENSITY, 0.2); 
	glFogf(GL_FOG_START, 0.1f); 
	glFogf(GL_FOG_END, 100.0f);
	tmp[0] = 0.8; tmp[1] = 1.0; tmp[2] = 0.9; tmp[3] = 1.0;
	glFogfv(GL_FOG_COLOR, tmp);
	glEnable(GL_FOG);  

	// modelview tranformation is just identity
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	resizeGL();
}


/**
 * drawGL: calls the OpenGL list to display the object and then 
 * 	commits the rendering scene.
 */
void TFOpenGLArea::drawGL() 
{
	//static GLfloat white[]={1.0, 1.0, 1.0, 1.0};
	//static GLfloat blue[]= {0.3,0.3,1.0,1.0};
	//static GLfloat green[]={0.3,1.0,0.3,1.0};
	//static GLfloat red[]=  {1.0,0.3,0.3,1.0};

	// start opengl rendering, you must put all opengl calls
	// inside make_current
	if (make_current())
		{
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		drawTerrain ();

		glFlush ();
		   
		swap_buffers();
		}
}


/*
 * button_press_event_impl
 */
gint TFOpenGLArea::button_press_event_impl(GdkEventButton *event) 
{
	return TRUE;
}


/*
 * key_press_event_impl
 */
gint TFOpenGLArea::key_press_event_impl(GdkEventKey *event)
{
	char 	key = event->keyval;
	int	i = p_HF->getWidth(),
		j = p_HF->getHeight();

	printf("Keypress was %d\n",key);
	switch (key) 
		{ 
		case 'x':  //GLUT_KEY_UP:
			d_meshRowStep += 1;
			if (d_meshRowStep >= (unsigned int)j)
				d_meshRowStep = j - 1;
			printf("d_meshRowStep = %d\n", d_meshRowStep);
			break;

		case 'y': //GLUT_KEY_DOWN:
			d_meshRowStep -= 1;
			if (d_meshRowStep <= 0)
				d_meshRowStep = 1;
			printf("d_meshRowStep = %d\n", d_meshRowStep);
			break;

		case 'z': //GLUT_KEY_RIGHT:
			d_meshColumnStep += 1;
			if (d_meshColumnStep >= (unsigned int)i)
				d_meshColumnStep = i - 1;
			printf("d_meshColumnStep = %d\n", d_meshColumnStep);
			break;

		case 'a': //GLUT_KEY_LEFT:
			d_meshColumnStep -= 1;
			if (d_meshColumnStep <= 0)
				d_meshColumnStep = 1;
			printf("d_meshColumnStep = %d\n", d_meshColumnStep);
			break;

		case 'h':
			printf("meshview help\n\n");
			printf("f            -  Filled mesh\n");
			printf("w            -  Wireframe mesh\n");
			printf("l            -  Toggle lighting\n");
			printf("n            -  Toggle normals\n");
			printf("c            -  Toggle colormap\n");
			printf("left arrow   -  Decrease column step\n");
			printf("right arrow  -  Increase column step\n");
			printf("down arrow   -  Decrease row step\n");
			printf("up arrow     -  Increase row step\n");
			printf("escape or q  -  Quit\n\n");
			break;

		case 'f':
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
			break;

		case 'w':
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
			break;

		case 'l':
		if (glIsEnabled(GL_LIGHTING))
			glDisable(GL_LIGHTING);
		else
			glEnable(GL_LIGHTING);
		break;

		case 'n':
			d_doDrawNormals = !d_doDrawNormals;
			break;

		case 'c':
			if (glIsEnabled(GL_COLOR_MATERIAL))
				glDisable(GL_COLOR_MATERIAL);
			else
				glEnable(GL_COLOR_MATERIAL);
			break;

		case 'q':
		case 27:
			break;
		}

	initGL ();
	return TRUE;
}

